# Copyright(c) 1994-2002 Inger Jorgensen
# Copyright(c) 2000-2002 Association of Universities for Research in Astronomy, Inc.

procedure imcoadd(images)

# Combine science images, includes cleaning for cosmic ray events
# This script was originally written for ground based optical CCD data
# It should work on any digital linear imaging data.
#
# First image in the list is used as reference image 'ref'
#
# if fl_find==yes daofind is used for identifying objects in 'ref'
#   else coordinate file 'ref'_pos
#
# if fl_map==yes the transformation is determined:
#    One or two common object(s) is/are marked on each image
#    Each transformation is determined interactively using geomap
# else name convention used for input files for transformation
#
# if fl_trn ==yes the transformation is made ==> 'image'_trn
# if fl_med ==yes  'ref' + SUM 'image'_trn = 'ref'_med
# if fl_add ==yes  'ref'_add = mean of masked images
# if fl_avg ==yes  'ref'_avg = mean of images, no masking
# 
# NAME CONVENTION
#
# 'ref'_pos     : daofind file for 'ref'
# 'ref'_cen     : centered positions from apphot.center
# 'ref'_mag.tab : STSDAS table with magnitudes used for scaling
# 'image'_pos   : approximate positions using x_offset,y_offset
# 'image'_cen   : centered positions from apphot.center
# 'image'_trn   : position input file for geomap, entry in database
# 'image'_trn   : transformed image
# 'ref'_med     : median filtered combined image of all input images
# 'ref'_add     : mean image cleaned for CR, scaled and weighted
#                 with relint if fl_scale=yes
# 'ref'_avg     : mean image, NOT cleaned for CR, scaled and weighted
#                 with relint if fl_scale=yes
#
# If you encounter unexplained problems, unlearn critical tasks
# geomap, geotran, all tasks related to apphot
#
# Version  Oct 11, 2000  IJ
#          May 15, 2001  IJ, MEF implementation
#          May 17, 2001  IJ, Fixpix all input files if DQ,immasks or
#                        badpix file available
#                        Use DQ as immasks if available (default yes)
#          May 21, 2001  IJ, really made it use DQ
#          May 31, 2001  IJ, changed gimverify call to use new version of gimverify
#          Jun  7, 2001  IJ, trapped failure to center objects
#                            trapped no objects in one or more of *_trn
#                            fixed rejection of all objects if rms=0.
#                            delete left over tmp images badpix.{fits,imh,hhh}
#          Jun  8, 2001  IJ, exit if fl_obj=no alignmethod!=header
#          Aug 30, 2001  IJ, changed badpix growth to prevent bus errors
#                            for large images
#                            fixed use of individual *.pl for ref image
#          Nov 1, 2001   BM, generalized for images with different gain/ron
#          Nov 6, 2001   BM, bug fix
#          Feb 28, 2002  IJ  v1.3 release
#  
# Reference: Jorgensen I., 2002, ApJS, submitted
#
# You may modify this script, in which case the comments in the
# final script should contain the two copyright lines in the start
# of the script.
#

char images {prompt="Images to be combined, first image as reference"}
char sci_ext {"SCI",prompt="Science extension name if MEF input"}
char var_ext {"VAR",prompt="Variance extension name if MEF input"}
char dq_ext {"DQ",prompt="Data quality extension name if MEF input"}
char immasks  {"DQ",prompt="Masks of bad pixels for each image"}
char database {"imcoadd.dat",prompt="Database file for transformations"}
real threshold {20.,prompt="Threshold for daofind, sigma above sky"}
real fwhm {7.,min=1,prompt="FWHMPSF for daofind, in pixels (ref)"}
real box {20.,min=5,prompt="Box size for centering (ref)"}
char alignmethod{"user",min="user|twodx|header",
   prompt="Method for rough alignment of images"}
char asection{"default",prompt="Image sections (2) for alignmethod=twodx"}
int  xwindow {181,min=11,prompt="Window size to search if using twodx"}
char key_xoff {"XOFFSET",prompt="Header keyword for instrument X-offset in mm"}
char key_yoff {"YOFFSET",prompt="Header keyword for instrument Y-offset in mm"}
real instscale {1.611444,min=0.000001,prompt="Scale for instrument offsets arcsec/mm"}
char sign_off {"negative",min="negative|positive",
   prompt="XOFFSET and YOFFSET sign relative to detector"}
char key_pixscale {"PIXSCALE",prompt="Header keyword for pixel scale in arcsec/pixel"}
real pixscale {1.,min=0.000001,prompt="Pixel scale arcsec/pixel"}
real dispmag {1.0,prompt="Magnify for display"}
bool rotate {no,prompt="Assume rotation > 0.5 deg"}
bool scale  {no,prompt="Allow significant scale differences"}
char geofitgeom{"rscale",
        min="shift|xyscale|rotate|rscale|rxyscale|general",
        prompt="Fitting geometry for geomap"}
int  order{3,min=2,prompt="Fitting order for geomap (geofitgeom=general)"}
real sigfit{2.5,min=0,prompt="Sigma rejection for geomap"}
int  niter{5,min=0,prompt="Iterations for geomap"}
real coolimit{0.3,min=0,prompt="Sigma in pixels for converged geomap"}
char geointer {"linear",min="nearest|linear|poly3|poly5|spline3",
        prompt="Interpolation for geotran"}
int  geonxblock {2048,prompt="geotran:X dim of working block size in pixels"}
int  geonyblock {2048,prompt="geotran:Y dim of working block size in pixels"}
char key_ron {"RDNOISE",prompt="Header keyword for read noise in electrons"}
char key_gain {"GAIN",prompt="Header keyqword for gain in electrons/ADU"}
real ron {1.,prompt="Read out noise [electrons]"}
real gain {1.,prompt="Gain [electrons/ADU]"}
real datamin {-1000.,prompt="Minimum data value below sky for good pixels"}
real datamax {50000.,prompt="Maximum data value for good pixels"}
real aperture {30.,prompt="Aperture radius for magnitudes for scaling"}
real limit {15.,prompt="Intensity limit for CR cleaning"}
bool key_limit {yes,prompt="Is limit given in sigmas above sky?"}
real lowsigma {7.,prompt="Sigma rejection limit for CRs"}
real lowlimit {500.,prompt="Absolute rejection limit for CRs"}
real scalenoise {0.,prompt="Noise term proportional to signal above sky"}
int  growthrad {1,min=0,max=1,prompt="Growth radius for CR, 0 or 1"}
char statsec {"default",prompt="Statistics section for sky determination"}
char badpixfile {"",prompt="Badpixel file or image"}
bool fl_inter {no,prompt="Inspect fits interactively"}
bool fl_refmark {no,prompt="Mark objects on display for ref image"}
bool fl_mark {no,prompt="Mark objects on display for all images"}
bool fl_fixpix {yes,prompt="Fixpix input images using DQ (MEF only)"}
bool fl_find {yes,prompt="Find objects with daofind"}
bool fl_map {yes,prompt="Determine the transformation"}
bool fl_trn {yes,prompt="Make transformation and subtract sky"}
bool fl_med {yes,prompt="Calculate median image"}
bool fl_add {yes,prompt="Calculate mean of masked images"}
bool fl_avg {yes,prompt="Calculate mean of images, no masking"}
bool fl_scale {yes,prompt="Combine mean images scaled with RELINT"}
bool fl_overwrite {no,prompt="Overwrite previous results"}
char logfile {"imcoadd.log",prompt="Logfile"}
bool verbose {yes,prompt="Verbose extended output"}
struct *inimag {prompt="List-directed struct, do not change"}
struct *inmask {prompt="List-directed struct, do not change"}
struct *input  {prompt="List-directed struct, do not change"}

begin

char n_ref, n_images, n_coords, n_database, n_realinim, n_inim
char l_sci_ext, l_var_ext, l_dq_ext, l_zero_ext, n_sci_ext
char n_statsec, n_sigim, n_mask, n_badpixfile, n_immasks, n_geointer
char n_geofitgeom, n_alignmethod, n_alignsec
int  n_xwindow
char n_key_xoff, n_key_yoff, n_sign_off, n_key_pixscale
real n_pixscale, n_instscale
real x0, y0, x11, y11, x21, y21, x12, y12, x22, y22, xc1, yc1, xc2, yc2
real x11_ref, y11_ref
real xsize1, ysize1, xsize2, ysize2
real x_offset, y_offset, n_dispmag
real n_thres, n_box, n_fwhm, l_ron, l_gain
real n_ron[500], n_gain[500], roneff, gaineff
char n_key_gain, n_key_ron
real n_lim, n_datamin, n_datamax, n_aperture, n_sigfit, n_coolimit
real n_lsig, n_llim, n_usig, n_ulim, n_scnoise
int  Xmax, Ymax, n_order, n_files, n_i, n_growthrad, n
int n_xsamp, n_ysamp, l_nxblock, l_nyblock
bool n_fl_find, n_fl_trn, n_fl_med, n_fl_map, n_fl_add, n_fl_scale
bool n_fl_msk, n_fl_avg, n_rotate, n_scale, n_fl_overwrite, n_fl_inter
bool l_fl_fixpix
bool l_verbose
char l_logfile

real n_medsky, n_insky, n_refint, n_inint, n_totint
real n_sigx, n_sigy, n_boxfac
int  n_lines, n_count, n_iter, n_ordoff, n_sec, n_ii
bool n_interact, n_break, n_badim, fl_obj, l_fl_mef, n_fl_del
char n_imtype, n_str, n_test
char n_correlation

char tmpcen, tmpcoo, tmpref, tmpnew, tmpim, tmplis, tmpout, tmpbad 
char tmpmli, tmpmag, n_m0, n_m1, n_m2, n_m3, n_m4, tmpba2, tmpdq
char tmpres, tmpin, n_tmpbase, tmplas
char tmpshift, tmpsh1, tmpsh2, tmpimlas
struct n_struct

cache("tstat","imgets","tinfo","gimverify")
l_logfile=logfile
if(l_logfile=="" || l_logfile==" ") {
  l_logfile="imcoadd.log"
  printlog("WARNING - IMCOADD: Using default logfile imcoadd.log",l_logfile,yes)
}
l_verbose = verbose 

printlog("-----------------------------------------------------------------------------",l_logfile,l_verbose)
date | scan(n_struct)
printlog("IMCOADD -- "//n_struct,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)

# make sure xray.ximages is NOT loaded as it contains another imcalc
if(defpac("ximages")) {
 printlog("ERROR - IMCOADD: Package ximages should be unloaded before running",l_logfile,yes)
 bye
}

# get output imagetype
show("imtype") | scan(n_imtype)
n_imtype="."//n_imtype

n_images = images ; n_immasks = immasks
l_sci_ext = "["//sci_ext//"]" ; l_var_ext = "["//var_ext//"]" 
l_dq_ext = "["//dq_ext//"]" ; l_zero_ext = "[0]"
n_sci_ext = sci_ext
n_database = database
n_dispmag = dispmag
n_thres = threshold ; n_box = box ; n_fwhm = fwhm ; n_rotate=rotate
n_scale = scale ; n_geofitgeom=geofitgeom ; n_alignmethod=alignmethod
n_alignsec = asection ; n_xwindow=xwindow
n_order = order ; n_sigfit=sigfit ; n_geointer=geointer ; n_iter=niter
n_coolimit=coolimit
l_fl_fixpix=fl_fixpix ; n_fl_find = fl_find
n_fl_map = fl_map ; n_fl_trn = fl_trn
n_fl_med = fl_med ; n_fl_add = fl_add ; n_fl_avg = fl_avg
n_fl_scale=fl_scale
n_fl_overwrite = fl_overwrite ; n_fl_inter=fl_inter
n_statsec = statsec ; n_badpixfile = badpixfile
n_key_gain = key_gain ; n_key_ron = key_ron
l_gain = gain ; l_ron = ron ; roneff=0.0 ; gaineff=0.0
n_lim = limit ; n_growthrad = growthrad ; n_scnoise=scalenoise
n_lsig = lowsigma ; n_llim=lowlimit 
n_usig = 500000. ; n_ulim=100000000.
n_datamin = datamin ; n_datamax = datamax ; n_aperture = aperture
 
# temporary list of images
tmplis = mktemp("tmplis")   # list of images to combine
tmpmli = mktemp("tmpmli")   # list of masks of images
tmpout = mktemp("tmpout")   # 
tmpbad = mktemp("tmpbad")   # badpixel file
tmpba2 = mktemp("tmpba2")   # badpixel file  no 2, used for input
tmpdq  = mktemp("tmpdq")    # Mask files from DQ (MEF only)
n_sigim = mktemp("tmpsig")  # sigma image for masking
n_mask = mktemp("tmpmsk")   # mask image
n_m0 = mktemp("tmpm0")      # mask image
n_m1 = mktemp("tmpm1")      # mask image
n_m2 = mktemp("tmpm2")      # mask image
n_m3 = mktemp("tmpm3")      # mask image
n_m4 = mktemp("tmpm4")      # mask image
tmpres = mktemp("tmpres")   # results from geomap, iterative run
tmpin  = mktemp("tmpin")    # filtered input for geomap, interative run

tmpcen = mktemp("tmpcen")
tmpcoo = mktemp("tmpcoo")
tmpref = mktemp("tmpref")
tmplas = mktemp("tmplas")   # used in mapping loop, only
tmpimlas = mktemp("tmpiml") # twodx
tmpshift = mktemp("tmpshi") # twodx
tmpsh1 = mktemp("tmpsh1")   # twodx
tmpsh2 = mktemp("tmpsh2")   # twodx
tmpnew = mktemp("tmpnew")
tmpim = mktemp("tmpim")
tmpmag = mktemp("tmpmag")

if(n_alignmethod!="user") {
  fl_mark=no ; fl_refmark=no
}

# n_images is either a list of images separated with commas or
# a list in a file in which case the name starts with @

# make the list of images into a list in a file
if(substr(n_image,1,1)!="@") 
  files(n_images,sort-, > tmplis)
else if (access(substr(n_images,2,strlen(n_images))))
 copy( (substr(n_images,2,strlen(n_images))),tmplis,verbose=no)
else {
 printlog("ERROR - IMCOADD: Input file "//substr(n_images,2,strlen(n_images))//" not found",l_logfile,yes)
 bye
}

# cut off get any extensions (.fits, .imh) on input images 
inimag=tmplis
while(fscan(inimag,n_inim)!=EOF) {
 if( substr(n_inim,strlen(n_inim)-4,strlen(n_inim))==".fits")
  n_inim=substr(n_inim,1,strlen(n_inim)-5)
 if( substr(n_inim,strlen(n_inim)-3,strlen(n_inim))==".imh")
  n_inim=substr(n_inim,1,strlen(n_inim)-4)
 print(n_inim, >> tmpmli)
}
inimag=""
delete(tmplis,verify-)
rename(tmpmli,tmplis,field="all")

# n_immasks is either a list of masks separated with commas,
# a list in a file in which case the name starts with @, or "DQ"

# make the list of immasks into a list in a file
if(n_immasks!="DQ") {
  if(substr(n_immasks,1,1)!="@")
    files(n_immasks,sort-, > tmpmli)
  else if (access(substr(n_immasks,2,strlen(n_immasks))))
   copy( (substr(n_immasks,2,strlen(n_immasks))),tmpmli,verbose=no)
  else {
   printlog("ERROR - IMCOADD: Input file "//substr(n_immasks,2,strlen(n_immasks))//" not found",l_logfile,yes)
   delete(tmplis,verify-)
   bye
  }
}

count(tmplis) | scan(n_files)
if(n_files < 2) {
  printlog("ERROR - IMCOADD: Only one image in list",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-, >> "dev$null")
  bye
} else if (n_files > 500) {
  printlog("ERROR - IMCOADD: The number of files exceeds the maximum allowable number of 500",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-, >> "dev$null")
  bye
}

# Check mask list
n_i=0 ; n_fl_msk = yes
if(access(tmpmli))
 count(tmpmli) | scan(n_i)
if(n_i==0)
 n_fl_msk=no
if(n_i>0 && n_i!=n_files) {
 printlog("ERROR - IMCOADD: Same number of images and masks required. Use none for no mask",l_logfile,yes)
 delete(tmplis//","//tmpmli,verify-, >> "dev$null")
 bye
}

printlog("Images (and masks) in list",l_logfile,l_verbose)
if(n_fl_msk)
  join(tmplis//","//tmpmli, >> l_logfile)
else
  type(tmplis, >> l_logfile)

if(l_verbose) {
if(n_fl_msk)
  join(tmplis//","//tmpmli)
else
  type(tmplis)
}

# Checking images and files
l_fl_mef = no  # Default is not MEF
inimag = tmplis
n_files = 0
while(fscan(inimag,n_inim) != EOF) {
    n_files += 1
  if(n_files == 1) 
    n_ref = n_inim

  gimverify(n_inim)
  if (gimverify.status==1) {
    printlog("ERROR - IMCOADD: Input image missing: "//n_inim,l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
    bye
  }
  if ((gimverify.status==0 && n_files>1 && !l_fl_mef) || \
      (gimverify.status>1 && n_files>1 && l_fl_mef) ) {
    printlog("ERROR - IMCOADD: Mix of MEF and other image formats",
      logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
    bye
  }
  if (gimverify.status==0) {
    l_fl_mef=yes
    if(!imaccess(n_inim//l_sci_ext) ) {
      printlog("ERROR - IMCOADD: Cannot access image : "//n_inim//"["//l_sci_ext//"]",
        l_logfile,yes)
      delete(tmplis//","//tmpmli,verify-)
      bye
    }
  } else {
    l_sci_ext="" ; l_var_ext="" ; l_dq_ext="" ; l_zero_ext=""
  }

  # ---- MEF: If DQ available: fixpix, use DQ as immasks
   if(l_fl_mef && imaccess(n_inim//l_dq_ext)) {
     if(l_fl_fixpix || n_fl_med) {
       if(n_files==1)
         printlog("Making individual masks from DQ",l_logfile,l_verbose)
       imexpr("(int(a/2.)*2!=a) ? 1 : 0",tmpdq//str(n_files)//".pl",
        n_inim//l_dq_ext,rangecheck+,verbose-)
     }
     if(l_fl_fixpix) {
       if(n_files==1)
         printlog("Fixpix input images using DQ",l_logfile,l_verbose)
       fixpix(n_inim//l_sci_ext,tmpdq//str(n_files)//".pl",linterp="",
        cinterp="",ninterp="1",creplace="",constant=0.)
       gemhedit(n_inim//l_zero_ext,"IMCOFIX","done","Imcoadd: fixpix using DQ")
     }

     if(n_immasks=="DQ") {
        print(tmpdq//str(n_files), >> tmpmli)
        n_fl_msk=yes
     }
   }

  # ----

  if(n_files==1 && n_fl_find && access(n_inim//"_pos") && !n_fl_overwrite ) {
    printlog("ERROR - IMCOADD: Output file "//n_inim//"_pos exists",l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
   bye
  } else if (n_files==1 && n_fl_find && access(n_inim//"_pos") && n_fl_overwrite ) 
    delete(n_inim//"_pos",verify-)

  if(n_files>1 && n_fl_map && access(n_inim//"_pos") && !n_fl_overwrite ) {
    printlog("ERROR - IMCOADD: Output file "//n_inim//"_pos exists",l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
   bye
  } else if (n_files>1 && n_fl_map && access(n_inim//"_pos") && n_fl_overwrite ) 
    delete(n_inim//"_pos",verify-)

  if(n_fl_map && (access(n_inim//"_cen") || access(n_inim//"_trn")) && !n_fl_overwrite ) {
    printlog("ERROR - IMCOADD: Output file(s) "//n_inim//"_cen and/or "//n_inim//"_trn exist",l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
   bye
  } else if (n_fl_map && (access(n_inim//"_cen") || access(n_inim//"_trn") ) && n_fl_overwrite) 
    delete(n_inim//"_cen,"//n_inim//"_trn",verify-, >>& "dev$null")

  if(n_fl_trn && imaccess(n_inim//"_trn") && !n_fl_overwrite ) {
    printlog("ERROR - IMCOADD: Output image "//n_inim//"_trn exists",l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
    bye
  } else if (n_fl_trn && imaccess(n_inim//"_trn") && n_fl_overwrite )
    imdelete(n_inim//"_trn",verify-)

  # Check ron and gain in image
  imgets(n_inim//l_zero_ext,n_key_ron, >>& "dev$null")
  if (imgets.value == "0") {
   printlog("WARNING - IMCOADD: Adopting default read noise of "//l_ron//" e-",
   l_logfile,yes)
   n_ron[n_files]=l_ron
  } else {
   n_ron[n_files]=real(imgets.value)
  }
  roneff=roneff+n_ron[n_files]**2

  imgets(n_inim//l_zero_ext,n_key_gain, >>& "dev$null")
  if (imgets.value == "0") {
   printlog("WARNING - IMCOADD: Adopting default gain of "//l_gain//" e-/ADU",
   l_logfile,yes)
   n_gain[n_files]=l_gain
  } else {
   n_gain[n_files]=real(imgets.value)
  }
  gaineff=gaineff+n_gain[n_files]
  #print(n_files,n_gain[n_files],n_ron[n_files],gaineff,roneff)

}
inimag = ""

# check for existing badpix files and existing magnitude output files
if(n_fl_med) {
  if(access(n_ref//"_mag.tab") && !n_fl_overwrite) {
     printlog("ERROR - IMCOADD: Output file "//n_ref//"_mag.tab exists",l_logfile,yes)
     delete(tmplis//","//tmpmli,verify-)
     bye
  } else if (access(n_ref//"_mag.tab") && n_fl_overwrite)
     delete(n_ref//"_mag.tab",verify-)

  inimag=tmplis
  while(fscan(inimag,n_inim)!=EOF) {
     if( imaccess(n_inim//"badpix.pl") && !n_fl_overwrite) {
       printlog("ERROR - IMCOADD: Bad pixel image "//n_inim//"badpix exists",l_logfile,yes)
       delete(tmplis//","//tmpmli,verify-)
       bye
     } else if ( imaccess(n_inim//"badpix.pl") && n_fl_overwrite)
       imdelete(n_inim//"badpix.pl",verify-)
     if ( imaccess(n_inim//"badpix") ) {
       printlog("WARNING - IMCOADD: Deleting left over temporary image "//n_inim//"badpix",
         l_logfile,yes)
       imdelete(n_inim//"badpix",verify-)
     }

     if (access(n_inim//"_trn_mag") && !n_fl_overwrite ) {
       printlog("ERROR - IMCOADD: Output file "//n_inim//"_trn_mag exists",l_logfile,yes)
       delete(tmplis//","//tmpmli,verify-)
       bye
     } else if ( access(n_inim//"_trn_mag") && n_fl_overwrite )
       delete(n_inim//"_trn_mag",verify-)

  }
  inimag=""
}

# Check access to badpix file
n_badim=no
if(n_badpixfile!="" && n_badpixfile!=" ") {
  if(!access(n_badpixfile) && !imaccess(n_badpixfile)) {
   printlog("WARNING - IMCOADD: Bad pixel file "//n_badpixfile//" not found",l_logfile,yes)
   n_badpixfile="gemtools$badpix.none"
  } else if(imaccess(n_badpixfile)) {
    n_badim=yes    # It's an image
  } else
    n_badim=no     # It's a file
} else
   n_badpixfile="gemtools$badpix.none"
printlog("Bad pixel file "//n_badpixfile,l_logfile,l_verbose)

printf("Read noise: %6.2f e-   Gain: %6.2f e-/ADU\n",n_ron[1],n_gain[1]) | scan(n_struct)
printlog(n_struct,l_logfile,l_verbose)

# cut one line off tmplis - this is the reference image
fields(tmplis,"1",lines="2-9999", > tmpout)
delete(tmplis,verify=no)
rename(tmpout,tmplis,field="root")

# Check access to median image
if(n_fl_med && imaccess(n_ref//"_med") && !n_fl_overwrite ) {
  printlog("ERROR - IMCOADD: Output image "//n_ref//"_med exists",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-)
  bye
} else if (n_fl_med && imaccess(n_ref//"_med") && n_fl_overwrite)
  imdelete(n_ref//"_med",verify-)

# Check access to cleaned average image
if(n_fl_add && imaccess(n_ref//"_add") && !n_fl_overwrite ) {
  printlog("ERROR - IMCOADD: Output image "//n_ref//"_add exists",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-)
  bye
} else if(n_fl_add && imaccess(n_ref//"_add") && n_fl_overwrite )
  imdelete(n_ref//"_add",verify-)

# Check access to uncleanded average image
if(n_fl_avg && imaccess(n_ref//"_avg") && !n_fl_overwrite ) {
  printlog("ERROR - IMCOADD: Output image "//n_ref//"_avg exists",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-)
  bye
} else if (n_fl_avg && imaccess(n_ref//"_avg") && n_fl_overwrite )
  imdelete(n_ref//"_avg",verify-)

# Check access to database if required
if(!n_fl_map && !access(n_database) && (n_fl_trn || n_fl_med) ) {
  printlog("ERROR - IMCOADD: Database file missing: "//n_database,l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-)
  bye
}

# Check access to median image if required
if(!n_fl_med && n_fl_add && !imaccess(n_ref//"_med")) {
  printlog("ERROR - IMCOADD: Median image "//n_ref//"_med does not exist",l_logfile,yes)
  printlog("                 Use fl_med=yes to create the median image",l_logfile,yes)
  delete(tmplis//","//tmpmli,verify-)
  bye
}

# check alignsec if needed
if(n_alignsec!="default" && n_fl_map && n_alignmethod=="twodx") {
  print(n_alignsec) | tokens("STDIN",ignore+,begin_com="#",
    end_comm="eol",newlines-) | \
    match(":","STDIN",stop-,print_file-,meta+) | count("STDIN") | scan(n_sec)
  if(n_sec!=4) {
    printlog("ERROR - IMCOADD: alignsec must be given as ",l_logfile,yes)
    printlog("       [x11:x12,y11:y12] [x21:x22,y21:y22] or set to default",l_logfile,yes)
    delete(tmplis//","//tmpmli,verify-)
    bye
  }
}

# check if xoffset,yoffset related stuff is needed
if(n_alignmethod=="header" && n_fl_map) {
  n_key_xoff=key_xoff ; n_key_yoff=key_yoff
  n_sign_off=sign_off ; n_key_pixscale=key_pixscale
  n_instscale=instscale
  if(n_rotate) {
    printlog("WARNING - IMCOADD: Rotation >0.5 deg not handled based on ",l_logfile,verbose+)
    printlog("                   xoffset and yoffset from header. Setting rotate=no",l_logfile,verbose+)
    n_rotate=no
  }

}

# size of image
imgets( n_ref//l_sci_ext, "i_naxis1" )
Xmax= int(imgets.value)
imgets( n_ref//l_sci_ext, "i_naxis2" )
Ymax= int(imgets.value)

# set statsec if default
if(n_statsec=="default") 
  n_statsec="[100:"//str(Xmax-100)//",100:"//str(Ymax-100)//"]"
printlog("Statistics section: "//n_statsec,l_logfile,l_verbose)

# Hardwire n_coords
n_coords=n_ref//"_pos"

# set fl_obj for a start
fl_obj=yes

# Find objects using daofind or check coordinate file existing
if(n_fl_find) {

# Get sky in order to estimate sigma of the sky - 
#  accurate value not required
imstat(n_ref//l_sci_ext//n_statsec,fields="midpt",format=no,
  lower=n_datamin,upper=n_datamax) | scan(n_medsky)

printlog("Finding objects in "//n_ref,l_logfile,l_verbose)
     daofind(n_ref//l_sci_ext,
     output=n_ref//"_pos", findpars.threshold=n_thres,
     fwhmpsf=n_fwhm, datapars.scale=1.,verify=no,interactive=no,
     findpars.roundlo=-1.,findpars.roundhi=1.,
     findpars.sharplo=0.2,findpars.sharphi=1.,findpars.nsigma=2.,
     datapars.datamin=n_datamin, datapars.datamax=n_datamax,
     datapars.sigma=(sqrt((n_ron[1]/n_gain[1])**2+n_medsky/n_gain[1])) )
     n_coords = n_ref//"_pos"
}
else {
  if(n_fl_map) {
    if(!access(n_coords)) {
      printlog("ERROR - IMCOADD: Coordinate file missing: "//n_coords,l_logfile,yes)
      bye
    }
  }
}

# Only some values work for sampling
n_xsamp  = 1 ; n_ysamp =1
if(Xmax>500)
  n_xsamp = 10
if(Ymax>500)
  n_ysamp = 10
if(Xmax>1000)
  n_xsamp = 20
if(Ymax>1000)
  n_ysamp = 20
l_nxblock=geonxblock ; l_nyblock=geonyblock
if(n_fl_trn) {
  printlog("Sampling for geotran  : "//n_xsamp//"  "//n_ysamp,l_logfile,l_verbose)
  printlog("Block size for geotran: "//l_nxblock//"  "//l_nyblock,l_logfile,l_verbose)
}

# start of mapping IF

if(n_fl_map) {
printlog("Centering objects in "//n_ref,l_logfile,l_verbose)
center( n_ref//l_sci_ext, coords=n_coords, output= n_ref//"_cen" , 
  interactive=no, 
  cbox=n_box, maxshift=n_box, centerpars.cthreshold=0.,
  datapars.scale=1., verify=no, datapars.datamin=n_datamin,
  datapars.datamax=n_datamax, centerpars.calgorithm="centroid" )

if(!access(n_ref//"_cen")) {
 n_test="0"
} else {
# Select only objects where first centering is OK
  txdump(n_ref//"_cen","xcenter,ycenter,cerror,id","cerror=='NoError'",
    headers=no) | match("No","STDIN",stop=no, > tmpref)

# --------------- changing settings to catch few objects -------------
# Settings to catch too few objects and/or n_iter=0
# Do not try to iterate if nobj<2, use n_iter=1 and no rejection
  n_test="10"
  count(tmpref) | scan(n_test)
}

# If no objects, the only working setting is
# n_iter=0, n_fl_scale=no, n_alignmethod="header"
fl_obj=yes  # there are objects
if(int(n_test)==0)  {
   printlog("WARNING - IMCOADD: No objects sucessfully centered",
    l_logfile,verbose+)
   if(n_alignmethod!="header") {
     printlog("ERROR - IMCOADD: Only alignmethod=header may work",
       l_logfile,verbose+)
     if(n_immasks=="DQ" && n_fl_msk) 
        imdelete("@"//tmpmli,verify=no, >>& "dev$null")
     delete(tmplis//","//tmpmli//","//tmpref,ver-, >>& "dev$null")
     bye
   } else {
     printlog("WARNING - IMCOADD: Setting niter=0 and fl_scale=no",
      l_logfile,verbose+)
     n_iter=0 ; n_fl_scale=no ; fl_obj=no
   }
}

if(int(n_test)<2 && n_iter>1) {
  printlog("WARNING - IMCOADD: Too few objects. Setting niter=1, sigfit=100",
     l_logfile,verbose+)
  n_iter=1
  n_sigfit=100
}
if(int(n_test)<6 && n_iter>2) {
  printlog("WARNING - IMCOADD: Too few objects. Setting niter=2",
     l_logfile,verbose+)
  n_iter=2
}

# Check the fit geometry depending on the number of objects, be safe
# shift|xyscale|rotate|rscale|rxyscale|general
if(int(n_test)==1) {
  if(n_geofitgeom!="shift")
    printlog("WARNING - IMCOADD: Too few objects. Setting geofitgeom=shift",
      l_logfile,verbose+)
  n_geofitgeom="shift"
  if(n_rotate) {
    printlog("WARNING - IMCOADD: Cannot determine rotation, setting rotate=no",
      l_logfile,verbose+)
    n_rotate=no
  }
}

if(int(n_test)==2 && n_geofitgeom!="shift" && n_geofitgeom!="rotate" && !n_rotate) {
  if(n_geofitgeom!="xyscale")
    printlog("WARNING - IMCOADD: Too few objects. Setting geofitgeom=xyscale",
      l_logfile,verbose+)
  n_geofitgeom="xyscale"
}

if(int(n_test)==2 && n_geofitgeom!="shift" && n_geofitgeom!="xyscale" && n_rotate) {
  if(n_geofitgeom!="rotate")
    printlog("WARNING - IMCOADD: Too few objects. Setting geofitgeom=rotate",
      l_logfile,verbose+)
  n_geofitgeom="rotate"
}

if(int(n_test)<6 && n_geofitgeom!="shift" && n_geofitgeom!="rotate" \
 && n_geofitgeom!="xyscale") {
  if(n_geofitgeom!="rscale")
    printlog("WARNING - IMCOADD: Too few objects. Setting geofitgeom=rscale",
      l_logfile,verbose+)
  n_geofitgeom="rscale"
}

if(n_iter==0)
  tmpres=n_database

printlog("Number of objects successfully centered in reference image "//n_test,
   l_logfile,l_verbose)

# --------------- end of changing settings ------------------

# set tmplas for the first round
tmplas=tmpref
# - - - - - - - 
printlog("Alignment method: "//n_alignmethod,l_logfile,l_verbose)

if(n_alignmethod=="user") {  # ---- start of user alignment for refim
print(" ==> Reference image: ", n_ref )
display( n_ref//l_sci_ext, 1, xmag=n_dispmag, ymag=n_dispmag )
if(fl_refmark) {
  print("Marking found objects in reference image")
fields(tmpref,fields="1,2,4",lines="1-",quit_if_miss=no,
  print_file_n=no) | tvmark(1,"STDIN",logfile="",autolog-,outimage="",
  deletions="",commands="",mark="point",color=202,pointsize=1,
  txsize=1,nxoffset=5,nyoffset=5,interactive=no,tolerance=1.5,
  interactive=no,label+,number-,font="raster")
}
if(fl_refmark)
  print("Objects found and centered are marked")

if (!n_rotate && !n_scale) {
  print("Point to one common object in reference image" )
  print("    strike any key")
  = fscan( imcur, x11, y11 )
} else {
  print("Point to first common object in reference image" )
  print("    strike any key")
  = fscan( imcur, x11, y11 )
  print("Point to second common object in reference image" )
  print("    strike any key")
  = fscan( imcur, x12, y12 )
}

} # ---- end of user alignment for reference image

if(n_alignmethod=="header") {  # ----- start of header alignment refim
  if(n_key_pixscale!="" && n_key_pixscale!=" ") {
    imgets(n_ref//l_zero_ext,n_key_pixscale, >>& "dev$null")
    if(imgets.value!="0")
      n_pixscale=real(imgets.value)
    else
      n_pixscale=pixscale
  } else
      n_pixscale=pixscale
  
  n_test=""
  hselect(n_ref//l_zero_ext,n_key_xoff,yes) | scan(n_test)
  imgets(n_ref//l_zero_ext,n_key_xoff, >>& "dev$null")
  if(imgets.value!="0" || n_test!="") {
     x11_ref=real(imgets.value)
     x11=Xmax/2
#     if(n_sign_off=="negative")
#       x11=-real(imgets.value)/n_pixscale+Xmax/2.
#     else
#       x11=real(imgets.value)/n_pixscale+Xmax/2.
  } else {
    printlog("ERROR - IMCOADD: XOFFSET not defined for image "//n_ref,
     l_logfile,verbose+)
    if(n_immasks=="DQ" && n_fl_msk) 
       imdelete("@"//tmpmli,verify=no, >>& "dev$null")
    delete(tmplis//","//tmpmli//","//tmpref,verify-)
    bye
  }
  n_test=""
  hselect(n_ref//l_zero_ext,n_key_yoff,yes) | scan(n_test)
  imgets(n_ref//l_zero_ext,n_key_yoff, >>& "dev$null")
  if(imgets.value!="0" || n_test!="") {
     y11_ref=real(imgets.value)
     y11=Ymax/2
#     if(n_sign_off=="negative")
#       y11=-real(imgets.value)/n_pixscale+Ymax/2.
#     else
#       y11=real(imgets.value)/n_pixscale+Ymax/2.
  } else {
    printlog("ERROR - IMCOADD: YOFFSET not defined for image "//n_ref,
     l_logfile,verbose+)
    if(n_immasks=="DQ" && n_fl_msk) 
       imdelete("@"//tmpmli,verify=no, >>& "dev$null")
    delete(tmplis//","//tmpmli//","//tmpref,verify-)
    bye
  }
} # ----- end of header alignment refim

if(n_alignmethod=="twodx") {
 tmpimlas = n_ref
 # setup sections for xregister
 if(n_alignsec=="default") {
   n_alignsec="["//int(0.25*Xmax)+1//":"//int(0.375*Xmax)//","//int(0.25*Ymax)+1//":"//int(0.375*Ymax)//"]"
   n_alignsec=n_alignsec//" ["//int(0.625*Xmax)+1//":"//int(0.75*Xmax)//","//int(0.625*Ymax)+1//":"//int(0.75*Ymax)//"]"
   xc1 = (int(0.25*Xmax)+1+int(0.375*Xmax))/2.
   yc1 = (int(0.25*Ymax)+1+int(0.375*Ymax))/2.
   xc2 = (int(0.625*Xmax)+1+int(0.75*Xmax))/2.
   yc2 = (int(0.625*Ymax)+1+int(0.75*Ymax))/2.
   xsize1=int(0.375*Xmax)-(int(0.25*Xmax)+1)
   ysize1=int(0.375*Ymax)-(int(0.25*Ymax)+1)
   xsize2=int(0.75*Xmax)-(int(0.625*Xmax)+1)
   ysize2=int(0.75*Ymax)-(int(0.625*Ymax)+1)
 } else {
  for(n_ii=1;n_ii<=4;n_ii+=1) {
  print(n_alignsec) | tokens("STDIN",ignore+,begin_com="#",
    end_comm="eol",newlines-) | \
    match(":","STDIN",stop-,print_file-,meta+) | \
    fields("STDIN","1",lines=str(n_ii),qui-,print-) | scan(n_str)
    x11=int( substr(n_str,1,stridx(":",n_str)-1) )
    x12=int( substr(n_str,stridx(":",n_str)+1,strlen(n_str)) )
    if(n_ii==1) {
       xc1=(x11+x12)/2. ; xsize1=x12-x11 }
    if(n_ii==2) {
       yc1=(x11+x12)/2. ; ysize1=x12-x11 }
    if(n_ii==3) {
       xc2=(x11+x12)/2. ; xsize2=x12-x11 }
    if(n_ii==4) {
       yc2=(x11+x12)/2. ; ysize2=x12-x11 }
  }
 } # end of if-else
 if( n_xwindow > min(xsize1,ysize1,xsize2,ysize2) ) {
   n_xwindow=min(xsize1,ysize1,xsize2,ysize2)
   printlog("WARNING - IMCOADD: Sections for X-correlation too small",l_logfile,yes)
   printlog("                   Resetting xwindow = "//str(n_xwindow),l_logfile,yes)
 } 
} # end of if
#print(xc1//" "//yc1//" "//xc2//" "//yc2)



# SCAN trough the other images
n=1
inimag = tmplis
while(fscan(inimag,n_inim) != EOF) {
n=n+1

if(n_alignmethod=="user") {  # ---- start of user alignment

print(" ==> Image to be transformed: ", n_inim )
display( n_inim//l_sci_ext, 1, xmag=n_dispmag, ymag=n_dispmag )

if(!n_rotate && !n_scale) {
  print("Point to one common object in image to be transformed" )
  print ("    coordinates for last image: ", x11, y11)
  print ("    strike any key")
  = fscan( imcur, x21, y21 )
  #x_offset = x21 - x11
  #y_offset = y21 - y11
  printf("%7.1f %7.1f %7.1f %7.1f\n",x11, y11,x21, y21, > tmpin)
  geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
   transforms=n_inim//"_trn",fitgeometry="shift",function="polynomial",
   verbose=l_verbose,interactive=no)
# shift the variables
if(n_iter>0) {
  x11 = x21 ; y11 = y21
}
#  count(tmpref) | scan(n_lines)
#  fields(tmpref,"1,2",lines="1-"//str(n_lines), > n_inim//"_pos")
#  tcalc(n_inim//"_pos","c1","c1 +"//str(x_offset),colfmt="f7.1")
#  tcalc(n_inim//"_pos","c2","c2 +"//str(y_offset),colfmt="f7.1")
} else {
  print("Point to first common object in image to be transformed" )
  print ("    coordinates for last image: ", x11, y11)
  print ("    strike any key")
  = fscan( imcur, x21, y21 )
  print("Point to second common object in image to be transformed" )
  print ("    coordinates for last image: ", x12, y12)
  print ("    strike any key")
  = fscan( imcur, x22, y22 )
# transform ref coordinates to next image's coordinates
# allow rotation and/or scale change
  printf("%7.1f %7.1f %7.1f %7.1f\n",x11, y11,x21, y21, > tmpin)
  printf("%7.1f %7.1f %7.1f %7.1f\n",x12, y12,x22, y22, >> tmpin)
  if(n_rotate && !n_scale) {
    geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
     transforms=n_inim//"_trn",fitgeometry="rotate",function="polynomial",
     verbose=l_verbose,interactive=no)
  } else if (n_scale) {
    geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
     transforms=n_inim//"_trn",fitgeometry="rscale",function="polynomial",
     verbose=l_verbose,interactive=no)
  }
# shift the variables
if(n_iter>0) {
  x11 = x21 ; y11 = y21
  x12 = x22 ; y12 = y22
}

}

}  # ---- end of user alignment

if(n_alignmethod=="header") {  # ----- start of header alignment 
  imgets(n_inim//l_zero_ext,n_key_pixscale)
  if(imgets.value!="0")
    n_pixscale=real(imgets.value)
  else
    n_pixscale=pixscale
  
  n_test=""
  hselect(n_inim//l_zero_ext,n_key_xoff,yes) | scan(n_test)
  imgets(n_inim//l_zero_ext,n_key_xoff, >>& "dev$null")
  if(imgets.value!="0" || n_test!="") {
     if(n_sign_off=="negative")
       x21=-(real(imgets.value)-x11_ref)*n_instscale/n_pixscale+Xmax/2.
     else
       x21=(real(imgets.value)-x11_ref)*n_instscale/n_pixscale+Xmax/2.
  } else {
    printlog("ERROR - IMCOADD: XOFFSET not defined for image "//n_inim,
     l_logfile,verbose+)
    inimag="" ; delete(tmplis//","//tmpmli//","//tmpref,verify-)
    bye
  }
  n_test=""
  hselect(n_inim//l_zero_ext,n_key_yoff,yes) | scan(n_test)
  imgets(n_inim//l_zero_ext,n_key_yoff, >>& "dev$null")
  if(imgets.value!="0" || n_test!="") {
     if(n_sign_off=="negative")
       y21=-(real(imgets.value)-y11_ref)*n_instscale/n_pixscale+Ymax/2.
     else
       y21=(real(imgets.value)-y11_ref)*n_instscale/n_pixscale+Ymax/2.
  } else {
    printlog("ERROR - IMCOADD: YOFFSET not defined for image "//n_inim,
     l_logfile,verbose+)
    inimag="" ; delete(tmplis//","//tmpmli//","//tmpref,verify-)
    bye
  }
  printf("%7.1f %7.1f %7.1f %7.1f\n",x11, y11,x21, y21, > tmpin)
  geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
   transforms=n_inim//"_trn",fitgeometry="shift",function="polynomial",
   verbose=l_verbose,interactive=no)
# shift the variables if niter>0
if(n_iter>0) {
  x11 = x21 ; y11 = y21
}

} # ----- end of header alignment

if(n_alignmethod=="twodx") { # --- start of twodx alignment
 printlog("xregister on the sections "//n_alignsec,l_logfile,l_verbose)
 printlog("Window size for search "//n_xwindow//" pixels",
   l_logfile,l_verbose)

 if(n_xwindow>21)
   n_correlation="fourier"
 else
   n_correlation="discrete"

 immatch.xregister(n_inim//l_sci_ext,tmpimlas//l_sci_ext,
   n_alignsec,tmpshift,databasefmt+,verbose=l_verbose,
   output="",append-,records="",coords="",xlag=0,ylag=0,dxlag=0,
   dylag=0,background="none",border=INDEF,loreject=INDEF,hireject=INDEF,
   apodize=0.,filter="none",correlation=n_correlation,
   xwindow=n_xwindow,ywindow=n_xwindow,function="centroid",
   xcbox=n_box,ycbox=n_box,
   interp_type = "linear",boundary_typ = "nearest",constant = 0.,
   interactive=no,graphics = "stdgraph",display = "stdimage",
   gcommands = "",icommands = "") | \
    tee(l_logfile,out_type="text",append+)

 printlog("",l_logfile,no)
 printlog("Output from xregister:",l_logfile,no)
 type(tmpshift, >> l_logfile)
 printlog("",l_logfile,no)

 fields(tmpshift,"3,4",lines="7", > tmpsh1)
 tcalc(tmpsh1,"c1","c1+"//str(xc1)) ; tcalc(tmpsh1,"c2","c2+"//str(yc1))
 tcalc(tmpsh1,"c3",str(xc1)) ; tcalc(tmpsh1,"c4",str(yc1))
 fields(tmpshift,"3,4",lines="8", > tmpsh2)
 tcalc(tmpsh2,"c1","c1+"//str(xc2)) ; tcalc(tmpsh2,"c2","c2+"//str(yc2))
 tcalc(tmpsh2,"c3",str(xc2)) ; tcalc(tmpsh2,"c4",str(yc2))
 fields(tmpsh1//","//tmpsh2,"1-4",lines="1-999", > tmpin)
 if(!n_rotate && !n_scale) {
  geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
   transforms=n_inim//"_trn",fitgeometry="shift",function="polynomial",
   verbose=l_verbose,interactive=no)
 }
 if(n_rotate && !n_scale) {
    geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
     transforms=n_inim//"_trn",fitgeometry="rotate",function="polynomial",
     verbose=l_verbose,interactive=no)
 } 
 if (n_scale) {
    geomap(tmpin,tmpres,xmin=1,xmax=Xmax,ymin=1,ymax=Ymax,
     transforms=n_inim//"_trn",fitgeometry="rscale",function="polynomial",
     verbose=l_verbose,interactive=no)
 }
 delete(tmpshift//","//tmpsh1//","//tmpsh2,verify-)
 if(n_iter>0) {
   tmpimlas=n_inim  # this image is reference for the next
 }
}  # ---- end of twodx aligment

# - - - - - - - - - 
delete(tmpin,verify=no)
if(fl_obj) {
  geoxytran(tmplas,n_inim//"_pos",tmpres,n_inim//"_trn",
    geometry="geometric",
    xref=INDEF,yref=INDEF,xmag=INDEF,xrotation=INDEF,yrotation=INDEF,
    xout=INDEF,yout=INDEF,xshift=INDEF,yshift=INDEF,xcol=1,ycol=2,
    calctype="real",xformat="",yformat="",min_sigdigit=7)
  if(tmpres!=n_database)
    delete(tmpres,verify=no, >>& "dev$null")
# Use this image as the next reference
  tmplas=n_inim//"_pos"

  printlog("Centering objects in "//n_inim,l_logfile,l_verbose)
  printlog("",l_logfile,l_verbose)
  center( n_inim//l_sci_ext, coords= n_inim//"_pos", \
  output= tmpcen , interactive=no, cbox=2.*n_box,
    maxshift=n_box, centerpars.cthreshold=0.,
    datapars.scale=1., verify=no, datapars.datamin=n_datamin,
    datapars.datamax=n_datamax )
  txdump(tmpcen,"xcenter,ycenter,id",yes,headers=no, > tmpcoo)
  center( n_inim//l_sci_ext, coords=tmpcoo, output= n_inim//"_cen" , interactive=no, 
    cbox=n_box, maxshift=n_box/1.5, centerpars.cthreshold=0.,
    datapars.scale=1., verify=no, datapars.datamin=n_datamin,
    datapars.datamax=n_datamax, centerpars.calgorithm="centroid" )
  delete(tmpcen//","//tmpcoo,verify=no)
  
  txdump(n_inim//"_cen","xcenter,ycenter,cerror",yes,headers=no, > tmpnew)
  # print("Ignore this warning")
  join   (tmpref, tmpnew) |
  match  ("NoError?*NoError", "STDIN", > tmpcoo)
  fields (tmpcoo, "1,2,5,6") | unique("STDIN", > n_inim//"_trn")

# Have to check that centering went ok
tinfo(n_inim//"_trn",ttout-)
if(tinfo.nrows==0 && n_iter>0) {
  printlog("ERROR - IMCOADD: Failed to center any objects",l_logfile,verbose+)
  printlog("                 Adjust datamax and box as needed. May have to use niter=0.",
     l_logfile,verbose+)
  printlog("                 Check "//n_inim//"_cen and "//n_inim//"_trn",
     l_logfile,verbose+)
  delete(tmpcoo//","//tmpnew//","//tmplis//","//tmpmli//","//tmpref,
     verify=no, >>& "dev$null")
  bye

} else {
  if(tmpres!=n_database)
    delete(tmpres,verify=no, >>& "dev$null")
}

if(fl_mark) {
print("Marking objects used for transformation")
fields(tmpcoo,fields="5,6,4",lines="1-",quit_if_miss=no,
  print_file_n=no) | tvmark(1,"STDIN",logfile="",autolog-,outimage="",
  deletions="",commands="",mark="point",color=202,pointsize=1,
  txsize=1,nxoffset=5,nyoffset=5,interactive=no,tolerance=1.5,
  interactive=no,label+,number-,font="raster")
}

delete(tmpcoo//","//tmpnew,verify=no, >>& "dev$null")

}

endcenter:  # jump point for failed centering  (is in fact not used anymore)

printf("Xmax = %5d   Ymax = %d\n",Xmax,Ymax) | scan(n_struct)
printlog(n_struct,l_logfile,l_verbose)
if(n_iter>0) {
  printlog("Entering geomap",l_logfile,l_verbose)
  printlog("Fitting geometry "//n_geofitgeom,l_logfile,l_verbose)
  if(n_geofitgeom=="general")
    printlog("Fitting order "//n_order,l_logfile,l_verbose)
  printlog("Iterating a maximum of "//str(n_iter)//" times",l_logfile,l_verbose)
} else 
  printlog("No iterations done with geomap",l_logfile,l_verbose)

# Iterate for geomap - geomap cannot do this on its own...
n_interact=no ; n_break=no ; n_tmpbase=tmpin
if(n_order>2 && n_iter>=3) 
  n_ordoff=n_order-2   # first 2 iterations with 2nd order fits
else
  n_ordoff=0

for(n_count=1;n_count<=n_iter;n_count+=1) {
  if(n_count>=3 || n_break)
    n_ordoff=0   # return to correct order
  if(n_count==n_iter || n_break) {
    n_interact=n_fl_inter ; n_tmpbase=n_database ; n_break=yes
    if(n_fl_inter) {
    print(" ") ; print("Last iteration, number "//n_count)
    }
  }
if((n_break || n_count==n_iter) && n_interact) {
print("---------------------------------------------------")
print(" x   graph x residuals ")
print(" y   graph y residuals ")
print(" d   delete nearest point ")
print(" u   undelete nearest point ")
print(" f   make new fit ")
print(" g   map data points ")
print(" q   exit curve fitting ")
print("---------------------------------------------------")
}

geomap(n_inim//"_trn",n_tmpbase,xmin=1,xmax=Xmax,ymin=1,ymax= Ymax,
  fitgeometry=n_geofitgeom,
  function="legendre",xxorder=(n_order-n_ordoff),
  xyorder=(n_order-n_ordoff),xxterms="half",yxorder=(n_order-n_ordoff),
  yyorder=(n_order-n_ordoff),yxterms="half", reject=n_sigfit, 
  interactive=n_interact,transforms=n_inim//"_trn",results=tmpres,
  verbose=l_verbose )
if(n_break)
  goto endgeoiter
# use these 4 lines if debugging with interactive=yes for all fits
# head(tmpres,nlines=5) | tokens(ignore-) | \
#   fields("STDIN","1",lines="29") | scan(n_sigx)
# head(tmpres,nlines=5) | tokens(ignore-) | \
#   fields("STDIN","1",lines="30") | scan(n_sigy)
#
head(tmpres,nlines=6) | tokens(ignore-) | \
  fields("STDIN","1",lines="37") | scan(n_sigx)
head(tmpres,nlines=6) | tokens(ignore-) | \
  fields("STDIN","1",lines="38") | scan(n_sigy)
delete(tmpin,verify=no, >>& "dev$null")
# stop when required accuracy is reached
if(n_sigx<=n_coolimit && n_sigy<=n_coolimit)
  n_break=yes

tselect(tmpres,tmpin,
  "abs(c7) <= "//str(n_sigx*n_sigfit)//" && abs(c8) <= "//str(n_sigy*n_sigfit))
count(tmpin) | scan(n_lines)
if(n_lines>=25) {
  delete(n_inim//"_trn",verify=no, >>& "dev$null")
  fields(tmpin,"1-4",lines="1-"//str(n_lines), > n_inim//"_trn")
} else
 n_break=yes

delete(tmpres//","//tmpin,verify-, >>& "dev$null")

}

# jump point for converged geomap
endgeoiter:
# delete last results file if there is such a thing
if(tmpres!=n_database)
  delete(tmpres,verify-, >>& "dev$null")

}  # End of SCAN through the other images
inimag = ""  # close file
#n_inim = ""  # close file
delete(tmpref,verify=no, >>& "dev$null")
}  # End of mapping IF

# start transformation IF
if(n_fl_trn) {

# Get sky for first image
imstat(n_ref//l_sci_ext//n_statsec,fields="midpt",format=no,
  lower=n_datamin,upper=n_datamax) | scan(n_medsky)
# changed to get better sky estimate  27.11.95
imstat(n_ref//l_sci_ext//n_statsec,fields="midpt",format=no,
  lower=max( n_datamin,n_medsky-5.*sqrt((n_ron[1]/n_gain[1])**2+n_medsky/n_gain[1]) ),
  upper=max( 50.,n_medsky+5.*sqrt((n_ron[1]/n_gain[1])**2+n_medsky/n_gain[1]) ) ) | scan(n_medsky)
printlog("",l_logfile,l_verbose)
printf("Median sky level for %s: %8.1f\n",n_ref,n_medsky) | scan(n_struct)
printlog(n_struct,l_logfile,l_verbose)
imarith(n_ref//l_sci_ext,"-",str(n_medsky),n_ref//"_trn",pixtype="real")
date | scan(n_struct)
gemhedit(n_ref//"_trn","GEM-TLM",n_struct,
   "Last modification with GEMINI")
gemhedit(n_ref//"_trn","IMCOADD",n_struct,"Time stamp for imcoadd")
gemhedit(n_ref//"_trn","MED_SKY",str(n_medsky),"Median sky level")

# SCAN trough the other images  - transform and subtract off sky
n=1
inimag = tmplis
while(fscan(inimag,n_inim) != EOF) {
n=n+1
printlog("Transforming "//n_inim//" to "//n_inim//"_trn",l_logfile,l_verbose)
# Check if input image is short !! 
  n_realinim = n_inim
  imgets(n_inim//l_sci_ext,"i_pixtype")
  if(imgets.value == "3") {
    chpixtype(n_inim//l_sci_ext,tmpim,"real",oldpixtype="all",verbose=no)
    n_realinim = tmpim
  }
  geotran( n_realinim//l_sci_ext, n_inim//"_trn", n_database, n_inim//"_trn", 
    xmin= 1, xmax= Xmax, ymin= 1, ymax= Ymax, 
    xsample=n_xsamp, ysample=n_ysamp, interpolant=n_geointer,
    nxblock=l_nxblock,nyblock=l_nyblock,
    geometry="geometric",xin=INDEF,yin=INDEF,xshift=INDEF,yshift=INDEF,
    xout=INDEF,yout=INDEF,xmag=INDEF,ymag=INDEF,xrotation=INDEF,
    yrotation=INDEF,xscale=1.,yscale=1., ncols=INDEF,nlines=INDEF,
    boundary="nearest",constant=0.,fluxconserve=yes,verbose=l_verbose )
  imdel(tmpim,verify=no, >>& "dev$null")

# sky level
  imstat(n_realinim//l_sci_ext//n_statsec,fields="midpt",format=no,
    lower=n_datamin,upper=n_datamax) | scan(n_insky)
# changed to get better sky estimate. 27.11.95
  imstat(n_realinim//l_sci_ext//n_statsec,fields="midpt",format=no,
    lower=max( n_datamin,n_insky-5.*sqrt((n_ron[n]/n_gain[n])**2+n_insky/n_gain[n]) ),
    upper=max( 50.,n_insky+5.*sqrt((n_ron[n]/n_gain[n])**2+n_insky/n_gain[n]) ) ) | scan(n_insky)
#  imstat(n_inim//n_statsec,fields="midpt",format=no,
#    lower=n_datamin,upper=(max(50.,3.*n_insky)) ) | scan(n_insky)
  printf("Median sky level for %s: %8.1f\n",n_inim,n_insky) | scan(n_struct)
  printlog(n_struct,l_logfile,l_verbose)

  imarith(n_inim//"_trn"//n_imtype,"-",str(n_insky),n_inim//"_trn"//n_imtype,
      pixtype="real")
  date | scan(n_struct)
  gemhedit(n_inim//"_trn","GEM-TLM",n_struct,
    "Last modification with GEMINI")
  gemhedit(n_inim//"_trn","IMCOADD",n_struct,"Time stamp for imcoadd")
  gemhedit(n_inim//"_trn","MED_SKY",str(n_insky),"Median sky level")

}    # end of SCAN trough the images
inimag = ""
}    # end of transformation IF

# combine to get median image IF
# first subtract backgrounds as midpt for each image
# the mean of the backgrounds is added to the median filtered image

if(n_fl_med) {

# make the badpixel mask for the first image, or copy it if already
# and image
if(!n_badim) {
# Is the badpixfile for trimmed or untrimmed image
type(n_badpixfile) | scan(tmpcen,tmpcoo)
printlog("",l_logfile,l_verbose)
printlog("Bad pixel file: "//n_badpixfile//" is for "//tmpcoo//" images",l_logfile,l_verbose)
if(tmpcoo == 'untrimmed') {
 imgets(n_ref,"TRIMSEC")
 x0 = real( substr(imgets.value,2,(stridx(":",imgets.value)-1) ) )
 imgets.value=substr(imgets.value,(stridx(",",imgets.value)+1),strlen(imgets.value) )
 y0 = real( substr(imgets.value,1,(stridx(":",imgets.value)-1) ) )
 fields(n_badpixfile,"1-4",lines="1-9999", > tmpbad)
 tcalc(tmpbad,"c1","c1+1-"//str(x0) )
 tcalc(tmpbad,"c2","c2+1-"//str(x0) )
 tcalc(tmpbad,"c3","c3+1-"//str(y0) )
 tcalc(tmpbad,"c4","c4+1-"//str(y0) )
 n_badpixfile=tmpbad
}
badpiximage(n_badpixfile,n_ref//l_sci_ext,n_ref//"badpix",goodvalue=0,
  badvalue=1)
delete(tmpbad,verify=no, >>& "dev$null")

} else 
imcopy(n_badpixfile,n_ref//"badpix.pl",verbose-)

# let tmpcen and tmpcoo have some resonable values again
tmpcen = mktemp("tmpcen")
tmpcoo = mktemp("tmpcoo")

# Get sky for first image from the header, 
#  write image to list of images to combine
imgets(n_ref//"_trn","med_sky") ; n_medsky=real(imgets.value)
printf("Median sky level for %s: %8.1f\n",n_ref,n_medsky) | scan(n_struct)
printlog(n_struct,l_logfile,l_verbose)
n_files = 1
print(n_ref//"_trn", > tmpout)

# put badpixel mask in first image
gemhedit(n_ref//"_trn","BPM",n_ref//"badpix.pl","Bad pixel mask")
# change pixel type to be able to transform this image 
chpixtype(n_ref//"badpix",n_ref//"badpix","real",
   oldpixtype="all",verbose=no)

# sky level for other images, transform their badpixel mask
inimag = tmplis
while(fscan(inimag,n_inim) != EOF) {

  imgets(n_inim//"_trn","med_sky") ; n_insky=real(imgets.value)
  printf("Median sky level for %s: %8.1f\n",n_inim,n_insky) | scan(n_struct)
  printlog(n_struct,l_logfile,l_verbose)
  n_medsky += n_insky 
  n_files += 1
  print(n_inim//"_trn", >> tmpout)   # list of images to combine

# make individual mask if required
  if(n_fl_msk) {
   n_fl_del=no
   fields(tmpmli,"1",lines=str(n_files)) | scan(tmpba2)
   if(tmpba2!="none" && imaccess(tmpba2) ) {
     n_mask=tmpba2
     imarith(n_mask,"+",n_ref//"badpix"//n_imtype,n_mask,calctype="real",
      pixtype="real",verbose-)
   } else if(tmpba2!="none" && access(tmpba2) ) { 
     print("Making individual mask for "//n_inim)
     type(tmpba2) | scan(tmpcen,tmpcoo)
     printlog("Mask file is for "//tmpcoo//" image",l_logfile,l_verbose)
     fields(tmpba2,"1-4",lines="1-9999", > tmpbad)
     if(tmpcoo == 'untrimmed') {
       imgets(n_inim,"TRIMSEC")
       x0 = real( substr(imgets.value,2,(stridx(":",imgets.value)-1) ) )
       imgets.value=substr(imgets.value,(stridx(",",imgets.value)+1),strlen(imgets.value) )
       y0 = real( substr(imgets.value,1,(stridx(":",imgets.value)-1) ) )
       tcalc(tmpbad,"c1","c1+1-"//str(x0) )
       tcalc(tmpbad,"c2","c2+1-"//str(x0) )
       tcalc(tmpbad,"c3","c3+1-"//str(y0) )
       tcalc(tmpbad,"c4","c4+1-"//str(y0) )
     }
     badpiximage(tmpbad,n_inim//l_sci_ext,n_mask,goodvalue=0,badvalue=1)
     delete(tmpbad,verify-, >>& "dev$null")
     imarith(n_mask,"+",n_ref//"badpix"//n_imtype,n_mask,calctype="real",
      pixtype="real")
   } else
     n_mask = n_ref//"badpix"

  } else
  n_mask = n_ref//"badpix"

# let tmpcen, tmpcoo, and tmpbad have some resonable values again
  tmpcen = mktemp("tmpcen")
  tmpcoo = mktemp("tmpcoo")
  tmpbad = mktemp("tmpbad")

# transform badpixelmask
if(n_mask!=n_ref//"badpix")  
  printlog("Transforming "//n_ref//"badpix +mask to "//n_inim//"badpix",l_logfile,l_verbose)
else
  printlog("Transforming "//n_ref//"badpix to "//n_inim//"badpix",l_logfile,l_verbose)
  geotran(n_mask, n_inim//"badpix", n_database, 
    n_inim//"_trn", xsample=n_xsamp, ysample=n_ysamp,  
    xmin= 1, xmax= Xmax, ymin= 1, ymax= Ymax, interpolant="linear",
    boundary="constant", constant=1.,
    nxblock=l_nxblock,nyblock=l_nyblock,
    geometry="geometric",xin=INDEF,yin=INDEF,xshift=INDEF,yshift=INDEF,
    xout=INDEF,yout=INDEF,xmag=INDEF,ymag=INDEF,xrotation=INDEF,
    yrotation=INDEF,xscale=1.,yscale=1., ncols=INDEF,nlines=INDEF,
    fluxconserve=yes,verbose=l_verbose )
#  imarith(n_inim//"badpix","+",0.75,n_inim//"badpix",pixtype="real")
#  chpixtype(n_inim//"badpix",n_inim//"badpix","int",
#    oldpixtype="all",verbose=no)
#  imcopy(n_inim//"badpix"//n_imtype,n_inim//"badpix.pl",verbose=no)
  imcalc(n_inim//"badpix"//n_imtype,n_inim//"badpix.pl","im1+0.75",
    pixtype="int", nullval=0., verbose=no)
  imdelete(n_inim//"badpix"//n_imtype,verify=no)

  gemhedit(n_inim//"_trn"//n_imtype,"BPM",n_inim//"badpix.pl",
    "Bad pixel mask")

if(n_mask!=n_ref//"badpix")          # delete mask if not n_refbadpix
  imdelete(n_mask,verify=no)

# let n_mask have tmp value again
n_mask = mktemp("tmpmsk")

}             # end of scan of inimag
inimag = ""
n_medsky = n_medsky/n_files

# make individual mask file for 'ref'
if(n_fl_msk) {
 fields(tmpmli,"1",lines="1") | scan(tmpba2)
   if(tmpba2!="none" && imaccess(tmpba2) ) {
     n_mask=tmpba2
    imarith(n_ref//"badpix","+",n_mask,n_ref//"badpix",calctype="real",
      pixtype="real",verbose-)
} else if(tmpba2!="none" && access(tmpba2) ) {
  print("Making individual mask for "//n_ref)
  type(tmpba2) | scan(tmpcen,tmpcoo)
  printlog("Mask pixel file is for "//tmpcoo//" image",l_logfile,l_verbose)
  fields(tmpba2,"1-4",lines="1-9999", > tmpbad)
  if(tmpcoo == 'untrimmed') {
    imgets(n_inim,"TRIMSEC")
    x0 = real( substr(imgets.value,2,(stridx(":",imgets.value)-1) ) )
    imgets.value=substr(imgets.value,(stridx(",",imgets.value)+1),strlen(imgets.value) )
    y0 = real( substr(imgets.value,1,(stridx(":",imgets.value)-1) ) )
    tcalc(tmpbad,"c1","c1+1-"//str(x0) )
    tcalc(tmpbad,"c2","c2+1-"//str(x0) )
    tcalc(tmpbad,"c3","c3+1-"//str(y0) )
    tcalc(tmpbad,"c4","c4+1-"//str(y0) )
  }
  badpiximage(tmpbad,n_inim,n_mask,goodvalue=0,badvalue=1)
  imarith(n_ref//"badpix","+",n_mask,n_ref//"badpix",calctype="real",
    pixtype="real")
  imdelete(n_mask,verify=no)
  delete(tmpbad,verify=no)
 }
}
# let tmpcen, tmpcoo and tmpbad have some resonable values again
tmpcen = mktemp("tmpcen")
tmpcoo = mktemp("tmpcoo")
tmpbad = mktemp("tmpbad")
n_mask = mktemp("tmpmsk")
# change pixtype back on ref badpix mask
# chpixtype(n_ref//"badpix",n_ref//"badpix","int",
#   oldpixtype="all",verbose=no)
imcopy(n_ref//"badpix"//n_imtype,n_ref//"badpix.pl",verbose=no)
imdelete(n_ref//"badpix"//n_imtype,verify=no)

#------------------------------------------------------
# check that there are objects in all *_trn
inimag=tmplis   # no suffix, ref image not included
while(fscan(inimag,n_inim)!=EOF) {
  tinfo(n_inim//"_trn",ttout-)
  if(tinfo.nrows==0)
    fl_obj=no
}
inimag=""

if(fl_obj) {  # ---- start of getting intensities of images from objects
inimag=tmplis   # no suffix, ref image not included
n_files = 1
while(fscan(inimag,n_inim) != EOF) {

n_files += 1

# make tables of _trn files and join using y-coordinate for
# objects in ref image  ==> one table

print("x_n1 r f6.2", > tmpbad)
print("y_n1 r f6.2", >> tmpbad)
print("x_n"//str(n_files)//" r f6.2", >> tmpbad)
print("y_n"//str(n_files)//" r f6.2", >> tmpbad)

if(n_files==2)
tcreate(tmpcoo,tmpbad,n_inim//"_trn",uparfile="",nskip=0,hist=no)
else {
  tcreate(tmpcen,tmpbad,n_inim//"_trn",uparfile="",nskip=0,hist=no)
#
  tjoin(tmpcoo,tmpcen,tmpnew,"y_n1","y_n1",tolerance=0.)
#
#  tmatch(tmpcoo,tmpcen,tmpnew,"x_n1,y_n1","x_n1,y_n1",
#    maxnorm=0.00000001)

  tdelete(tmpcoo//","//tmpcen,verify=no)
  tchcol(tmpnew,"x_n1_1","x_n1","","",verbose=no)
  tchcol(tmpnew,"x_n1_2","junk","","",verbose=no)
# this is the fix!
#  tproject(tmpnew,tmpcoo,"x*,y*",uniq=yes)
  tproject(tmpnew,tmpcoo,"x_n1,y_n1",uniq=yes)
  tdelete(tmpnew,verify=no)
}
delete(tmpbad,verify=no)

}
inimag=""   # close file

# get magnitudes and make one table with output
# coordinates are always as for ref image because we use
# the transformed sky subtracted images
tdump(tmpcoo,cdfile="STDOUT",pfile="STDOUT",datafile=tmpcen,
  columns="x_n1,y_n1",rows="-", >>& "dev$null")

for(n_i=1 ; n_i<=n_files ; n_i+=1) {
# tmpout is list with suffix  _trn
fields(tmpout,"1",lines=str(n_i)) | scan(n_inim)
phot(n_inim//""//n_imtype,datamin=n_datamin,datamax=n_datamax,
  ccdread="",gain="",exposure="",airmass="",filter="",obstime="",
  readnoise=n_ron[n_i],epadu=n_gain[n_i],itime=1.,calgorithm="none",
  xairmass=INDEF,ifilter="INDEF",otime="INDEF",salgorithm="constant",
  skyvalue=0.,apertures=str(n_aperture),zmag=22.,coords=tmpcen,
  output=n_inim//"_mag",interactive=no,verify=no)

print("x_n"//str(n_i)//" r f6.2", > tmpbad)
print("y_n"//str(n_i)//" r f6.2", >> tmpbad)
print("mag_n"//str(n_i)//" r f6.2", >> tmpbad)
print("perr_n"//str(n_i)//" ch*15 a", >> tmpbad)

if(n_i==1) {
  txdump(n_inim//"_mag","xcenter,ycenter,mag,perror",yes,headers=no) | \
   tcreate(tmpmag,tmpbad,"STDIN",uparfile="",nskip=0,hist=no)
  delete(tmpbad,verify=no)
} else {
  txdump(n_inim//"_mag","xcenter,ycenter,mag,perror",yes,headers=no) | \
   tcreate(tmpnew,tmpbad,"STDIN",uparfile="",nskip=0,hist=no)
  delete(tmpbad,verify=no)
  tmerge(tmpmag//","//tmpnew,tmpbad,"merge")
  tdelete(tmpmag//","//tmpnew,verify=no)
  rename(tmpbad//".tab",tmpmag//".tab")
}

}
delete(tmpcen,verify=no)
tdelete(tmpcoo,verify=no)

# select stars for which everything is ok
for(n_i=1 ; n_i<=n_files ; n_i+=1) {
  tselect(tmpmag,tmpnew,
   "mag_n"//str(n_i)//"<9999. && perr_n"//str(n_i)//"=='NoError'")
  tdelete(tmpmag,verify=no)
  rename(tmpnew//".tab",tmpmag//".tab")
}

# check output table one last time in case all objects are bad
tinfo(tmpmag//".tab",ttout-, >>& "dev$null")
if(tinfo.nrows==0) {
  printlog("WARNING - IMCOADD: No good objects in common. Setting fl_scale=no",l_logfile,verbose+)
  delete(tmpmag//".tab",verify-, >>& "dev$null")
  n_fl_scale=no
  goto noscaling
}

printlog("",l_logfile,l_verbose)
# Calculate mean intensity
n_totint=0.
for(n_i=1 ; n_i<=n_files ; n_i+=1) {
  fields(tmpout,"1",lines=str(n_i)) | scan(n_inim)  # image name
  tcalc(tmpmag,"tot_"//str(n_i),"10**(0.4*(22.-mag_n"//str(n_i)//"))" )
  tstat(tmpmag,"tot_"//str(n_i),outtable="STDOUT",rows="-", 
    >>& "dev$null")
  if(n_i==1)
    n_refint = tstat.mean
  n_inint = tstat.mean
  printf("Mean intensity for   %s:  (N,absolute,relative)= %4d %8.0f %6.3f\n",
     n_inim,tstat.nrows,n_inint,(n_inint/n_refint) ) | scan(n_struct)
  printlog(n_struct,l_logfile,l_verbose)
# put the relative intensity in the image header
# the images are *multiplied* by this factor, so it has to be the
# inverse relative intensity
  gemhedit(n_inim//""//n_imtype,"RELINT",(n_refint/n_inint),
    "Inverse of the relative intensity")
}
# save the magnitudes for control
rename(tmpmag//".tab",n_ref//"_mag.tab")

} else { # ---- end of getting intensities of images from objects
# if no objects, turn off scaling
  if(n_fl_scale) {
     n_fl_scale=no
     printlog("WARNING - IMCOADD: No objects. Setting fl_scale=no",l_logfile,verbose+)
  }
}
#------------------------------------------------------
noscaling:

# threshold not used for combining!
# the median image is on the scale of the reference image

if(n_fl_scale==yes) {
imcombine("@"//tmpout,n_ref//"_med",combine="median",reject="none",
  outtype="real",offsets="none",masktype="goodvalue",maskvalue=0,
  scale="!RELINT",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec="",logfile=l_logfile)
  date | scan(n_struct)
  gemhedit(n_ref//"_med","GEM-TLM",n_struct,
    "Last modification with GEMINI")
  gemhedit(n_ref//"_med","IMCOADD",n_struct,"Time stamp for imcoadd")
  hedit(n_ref//"_med","COMBSC","imcoadd combine scaled with RELINT",
    add=yes,verify=no,show=no,update=yes,delete=no)
} else {
imcombine("@"//tmpout,n_ref//"_med",combine="median",reject="none",
  outtype="real",offsets="none",masktype="goodvalue",maskvalue=0,
  scale="none",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec="",logfile=l_logfile)
  date | scan(n_struct)
  gemhedit(n_ref//"_med","GEM-TLM",n_struct,
    "Last modification with GEMINI")
  gemhedit(n_ref//"_med","IMCOADD",n_struct,"Time stamp for imcoadd")
  hedit(n_ref//"_med","COMBSC","imcoadd combine not scaled with RELINT",add=yes,
    verify=no,show=no,update=yes,delete=no)
}

gemhedit(n_ref//"_med","med_sky",str(n_medsky),"Median sky level")
# remove reference to BPM for reference image
hedit(n_ref//"_med","BPM",add=no,delete=yes,verify=no,show=no,update=yes)

# add sky to median image if cleaned mean images not to be made
if(!n_fl_add) {
imarith(n_ref//"_med","+",str(n_medsky),n_ref//"_med",
  pixtype="real")
}

# Update gain and readnoise keywords
imgets(n_ref//"_med","GAINORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_med","GAINORIG",n_gain[1],"Input gain")
}
imgets(n_ref//"_med","RONORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_med","RONORIG",n_ron[1],"Input read-noise")
}
# Effective ron and gain for combined image
# ron in e-/pix  as expected by other iraf tasks
if(n_key_gain=="" || n_key_gain==" ")
  n_key_gain="GAIN"
gemhedit(n_ref//"_med",n_key_gain,(2.*gaineff/3.),"Effective gain [e-/ADU]")

if(n_key_ron=="" || n_key_ron==" ")
  n_key_ron="RDNOISE"
gemhedit(n_ref//"_med",n_key_ron,(sqrt(2.*roneff/3.)),"Effective read-noise [e-]")

}  # end of fl_med IF

# check if n_fl_add=yes and n_fl_med=no
# if so assume that sky has to be subtracted off median image
if(n_fl_add && !n_fl_med) {
 print("Subtracting sky level off median image")
 imgets(n_ref//"_med","med_sky")
 imarith(n_ref//"_med","-",real(imgets.value),n_ref//"_med")

} 

# make the list of images if not made during fl_med
# get the median sky -- not safe to get it from median image
# as it may not exist if fl_med+ was never run
if(!n_fl_med && (n_fl_add || n_fl_avg)) {
  print(n_ref//"_trn", > tmpout)
  imgets(n_ref//"_trn","med_sky")
  n_medsky=real(imgets.value) ; n_files=1
  inimag = tmplis
    while(fscan(inimag,n_inim) != EOF) {
      print(n_inim//"_trn", >> tmpout)   # list of images to combine
      imgets(n_inim//"_trn","med_sky")
      n_medsky=n_medsky+real(imgets.value) ; n_files+=1
    }
  inimag=""
n_medsky=n_medsky/n_files
print("Median sky: "//str(n_medsky))
}


# make masks for individual images by comparison to median image
# then calculate mean of all images
if(n_fl_add) {

printlog("",l_logfile,l_verbose)

n=0
inimag = tmpout   # suffix _trn
while(fscan(inimag,n_inim) != EOF) {
  n=n+1

  n_ron[n] = (n_ron[n]/n_gain[n])**2          # RON^2 in ADU
  n_scnoise = n_scnoise**2           # Noise scaling ^2  (fraction)
  # make n_lim absolute if key_limit=yes
  if(key_limit)
    n_lim=n_lim*sqrt(n_ron[n]+n_medsky/n_gain[n])

  printlog("Masking cosmic ray events in "//n_inim,l_logfile,l_verbose)
# difference image, get the scale, remember the header contains
# the inverse scale
  if(n_fl_scale) {
     imgets(n_inim,"RELINT") ; n_inint = 1./real(imgets.value)
  } else 
     n_inint = 1.

  if (imaccess(n_inim//"_dif") && n_fl_overwrite) {
	imdelete(n_inim//"_dif", verify-)
  }
  imcalc(n_inim//","//n_ref//"_med",n_inim//"_dif",
   "im1/"//str(n_inint)//"-im2",pixtype="real",
   nullval=0.,verbose=no)
# sigma image, get sky value
  imgets(n_inim,"MED_SKY") ; n_insky = real(imgets.value)

print("sqrt(( 1./("//str(n_inint)//"**2)+\
1.5625/"//str(n_files)//")*"//str(n_ron[n])//" + \
( (im1+"//str(n_insky)//")/("//str(n_inint)//"**2) + \
(im2+"//str(n_medsky)//")*1.5625/"//str(n_files)//" ) /"//str(n_gain[n])//\
" + im2**2*"//n_scnoise//" )", 
> tmpbad)

  imcalc(n_inim//","//n_ref//"_med",n_sigim,"@"//tmpbad,pixtype="real",
    nullval=0., verbose=no)
  delete(tmpbad,verify=no)

# mask image  one: deviation, positive only
#            zero: no deviation
print("(im3<="//n_lim//" && (im1 >"//n_lsig//"*im2 || im1 >"//n_llim//"))",
 > tmpbad)
# || \
# (im3>"//n_lim//" && (im1 >"//n_usig//"*im2 || im1 >"//n_ulim//"))",
#  > tmpbad )

  imcalc(n_inim//"_dif,"//n_sigim//","//n_ref//"_med",n_mask//".pl",
    "@"//tmpbad,pixtype="int", nullval=0., verbose=no)

delete(tmpbad,verify=no)
# if growthrad=1 shift n_mask around and add
if(n_growthrad>0) {
  printlog("Applying growth radius "//n_growthrad,l_logfile,l_verbose)

# Code changed Aug 30, 2001 to solve bus errors for large images
# Make mask 1 and 0 only
  imcalc(n_mask//".pl",n_m4//".pl","if im1>0 then 1 else 0",verbose-)
# Make a mask with huge values in badpix
  imarith(n_m4//".pl","*","1000.",n_m0//".pl")
# Combine and use output plfile as the growth
  imcombine(n_mask//".pl,"//n_m0//".pl",n_m1//".pl",plfile=n_m2//".pl",
    combine="average",reject="minmax",nlow=0,nhigh=1,grow=1.,blank=1,
    logfile="",project=no,outtype="real",offsets="none",masktype="none",        
    scale="none",zero="none",weight="none",statsec="",expname="",
    lthreshold=INDEF,hthreshold=INDEF,nkeep=1)
# Add the original mask and the growth   
  imcalc(n_mask//".pl,"//n_m2//".pl",n_m3//".pl",
      "if im1>0 || im2>1 then 1 else 0", verbose-)
# Clean up
  imdelete(n_mask//".pl,"//n_m0//".pl,"//n_m1//".pl,"//n_m2//".pl,"//n_m4//".pl",ver-)
  imrename(n_m3//".pl",n_mask//".pl",verbose-)

##  imrename(n_mask//".pl",n_m0//".pl",verbose=no)
##  imcopy(n_m0//".pl",n_m1//".pl",verbose=no)
##  imcopy(n_m0//".pl[2:"//str(Xmax)//",1:"//str(Ymax-1)//"]",
##    n_m1//".pl[1:"//str(Xmax-1)//",1:"//str(Ymax-1)//"]",
##    verbose=no)
##  imcopy(n_m0//".pl",n_m2//".pl",verbose=no)
##  imcopy(n_m0//".pl[1:"//str(Xmax-1)//",1:"//str(Ymax-1)//"]",
##    n_m2//".pl[2:"//str(Xmax)//",1:"//str(Ymax-1)//"]",
##    verbose=no)
##  imcopy(n_m0//".pl",n_m3//".pl",verbose=no)
##  imcopy(n_m0//".pl[1:"//str(Xmax-1)//",2:"//str(Ymax)//"]",
##    n_m3//".pl[1:"//str(Xmax-1)//",1:"//str(Ymax-1)//"]",
##    verbose=no)
##  imcopy(n_m0//".pl",n_m4//".pl",verbose=no)
##  imcopy(n_m0//".pl[1:"//str(Xmax-1)//",1:"//str(Ymax-1)//"]",
##    n_m4//".pl[1:"//str(Xmax-1)//",2:"//str(Ymax)//"]",
##    verbose=no)
##  print(n_m0//".pl", > tmpbad) ; print(n_m1//".pl", >> tmpbad)
##  print(n_m2//".pl", >> tmpbad) ; print(n_m3//".pl", >> tmpbad)
##  print(n_m4//".pl", >> tmpbad)
###  type(tmpbad)
##  imcalc("@"//tmpbad,n_mask//".pl",
##    "((im1+im2+im3+im4+im5) > 0)",pixtype="int", 
##    nullval=0., verbose=no)
##  imdelete("@"//tmpbad,verify=no)
##  delete(tmpbad,verify=no)
}


# add the mask image to the badpixmask for the image
n_inim = substr( n_inim,1,(strlen(n_inim)-4) )
   imarith(n_inim//"badpix.pl","+",n_mask//".pl",n_inim//"badpix.pl",
     pixtype="int")

# delete  mask and sigma and difference
   imdelete(n_mask//".pl,"//n_sigim//","//n_inim//"_trn_dif",verify=no)

}
inimag = ""  # close file, end of masking

# combine the images with masking

# threshold not used for combining, scale if fl_scale=yes
if(n_fl_scale) {
imcombine("@"//tmpout,n_ref//"_add",combine="average",reject="none",
  outtype="real",offsets="none",masktype="goodvalue",maskvalue=0,
  scale="!RELINT",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec=n_statsec,logfile=l_logfile)
date | scan(n_struct)
#gemhedit(n_ref//"_add","GEM-TLM",n_struct,
#  "Last modification with GEMINI")
#gemhedit(n_ref//"_add","IMCOADD",n_struct,"Time stamp for imcoadd")
#gemhedit(n_ref//"_add","IMCOBPM",n_badpixfile,"Detector bad pixel file")
#hedit(n_ref//"_add","COMBSC","imcoadd combine scaled with RELINT",add=yes,
#  verify=no,show=no,update=yes,delete=no)
} else {
imcombine("@"//tmpout,n_ref//"_add",combine="average",reject="none",
  outtype="real",offsets="none",masktype="goodvalue",maskvalue=0,
  scale="none",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec=n_statsec,logfile=l_logfile)
date | scan(n_struct)
#gemhedit(n_ref//"_add","GEM-TLM",n_struct,
#  "Last modification with GEMINI")
#gemhedit(n_ref//"_add","IMCOADD",n_struct,"Time stamp for imcoadd")
#gemhedit(n_ref//"_add","IMCOBPM",n_badpixfile,"Detector bad pixel file")
#hedit(n_ref//"_add","COMBSC","imcoadd combine not scaled with RELINT",add=yes,
#  verify=no,show=no,update=yes,delete=no)
}

# Add median sky to mean image
imarith(n_ref//"_add","+",str(n_medsky),n_ref//"_add",
  pixtype="real")

# Pack up as MEF if input is MEF
if(l_fl_mef) {
 imrename(n_ref//"_add",tmpimlas,verbose-)
 wmef(tmpimlas,n_ref//"_add",extnames=n_sci_ext,phu=n_ref,verbose-)
 imdelete(tmpimlas,verify-)
}

gemhedit(n_ref//"_add"//l_zero_ext,"GEM-TLM",n_struct,
  "Last modification with GEMINI")
gemhedit(n_ref//"_add"//l_zero_ext,"IMCOADD",n_struct,"Time stamp for imcoadd")
gemhedit(n_ref//"_add"//l_zero_ext,"IMCOBPM",n_badpixfile,"Detector bad pixel file")

if(n_fl_scale) {
hedit(n_ref//"_add"//l_zero_ext,"COMBSC","imcoadd combine scaled with RELINT",add=yes,
  verify=no,show=no,update=yes,delete=no)
} else {
hedit(n_ref//"_add"//l_zero_ext,"COMBSC","imcoadd combine not scaled with RELINT",add=yes,
  verify=no,show=no,update=yes,delete=no)

}

# Add median sky  to median image 
imarith(n_ref//"_med","+",str(n_medsky),n_ref//"_med",
  pixtype="real")

gemhedit(n_ref//"_add"//l_zero_ext,"med_sky",str(n_medsky),"Median sky level")

# various header info in n_ref//"_add"

# Input ron and gain
# Update gain and readnoise keywords
imgets(n_ref//"_add"//l_zero_ext,"GAINORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_add"//l_zero_ext,"GAINORIG",n_gain[1],"Input gain")
   gemhedit(n_ref//"_add"//l_sci_ext,"GAINORIG",n_gain[1],"Input gain")
}
imgets(n_ref//"_add"//l_zero_ext,"RONORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_add"//l_zero_ext,"RONORIG",(sqrt(n_ron[1])*n_gain[1]),"Input read-noise")
   gemhedit(n_ref//"_add"//l_sci_ext,"RONORIG",(sqrt(n_ron[1])*n_gain[1]),"Input read-noise")
}

# Effective ron and gain for combined image
# ron in e-/pix  as expected by other iraf tasks
if(n_key_gain=="" || n_key_gain==" ")
  n_key_gain="GAIN"
gemhedit(n_ref//"_add"//l_zero_ext,n_key_gain,gaineff,"Effective gain [e-/ADU]")
gemhedit(n_ref//"_add"//l_sci_ext,n_key_gain,gaineff,"Effective gain [e-/ADU]")

if(n_key_ron=="" || n_key_ron==" ")
  n_key_ron="RDNOISE"
gemhedit(n_ref//"_add"//l_zero_ext,n_key_ron,(sqrt(roneff)),"Effective read-noise [e-]")
gemhedit(n_ref//"_add"//l_sci_ext,n_key_ron,(sqrt(roneff)),"Effective read-noise [e-]")

#if(n_key_ron=="" || n_key_ron==" ")
#  n_key_ron="RDNOISE"
#gemhedit(n_ref//"_add"//l_zero_ext,n_key_ron,str(sqrt(n_files)*sqrt(n_ron)*n_gain),
#  "Effective read-noise [e-]")

#if(n_key_gain=="" || n_key_gain==" ")
#  n_key_gain="GAIN"
#gemhedit(n_ref//"_add"//l_zero_ext,n_key_gain,str(n_files*n_gain),
#  "Effective gain [e-/ADU]")

gemhedit(n_ref//"_add"//l_zero_ext,"SCNOISE",str(sqrt(n_scnoise)),
  "imcoadd scnoise")
gemhedit(n_ref//"_add"//l_zero_ext,"LIMIT",str(n_lim),"imcoadd limit for cleaning")
gemhedit(n_ref//"_add"//l_zero_ext,"LOWSIG",str(n_lsig),"imcoadd sigma rejection")
gemhedit(n_ref//"_add"//l_zero_ext,"LOWLIM",str(n_llim),
  "imcoadd absolute rejection limit")
gemhedit(n_ref//"_add"//l_zero_ext,"GROWTHR",str(n_growthrad),
  "imcoadd growth radius")
gemhedit(n_ref//"_add"//l_zero_ext,"GEOGEOM",n_geofitgeom,
  "imcoadd geomap geometry")
gemhedit(n_ref//"_add"//l_zero_ext,"GEOINTER",n_geointer,
  "imcoadd geotrans interpolation")
# remove reference to BPM for reference image
hedit(n_ref//"_add"//l_sci_ext,"BPM",delete=yes,verify=no,show=no,update=yes,add=no)

}  # end of fl_add

# uncleaned mean image - fl_avg=yes
if(n_fl_avg) {
if(n_fl_scale) {
imcombine("@"//tmpout,n_ref//"_avg",combine="average",reject="none",
  outtype="real",offsets="none",masktype="none",maskvalue=0,
  scale="!RELINT",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec=n_statsec,logfile=l_logfile)
date | scan(n_struct)
gemhedit(n_ref//"_avg","GEM-TLM",n_struct,
   "Last modification with GEMINI")
gemhedit(n_ref//"_avg","IMCOADD",n_struct,"Time stamp for imcoadd")
hedit(n_ref//"_avg","COMBSC","imcoadd combine scaled with RELINT",add=yes,
  verify=no,show=no,update=yes,delete=no)
} else {
imcombine("@"//tmpout,n_ref//"_avg",combine="average",reject="none",
  outtype="real",offsets="none",masktype="none",maskvalue=0,
  scale="none",lthreshold=INDEF,hthreshold=INDEF,expname="",
  zero="none",weight="none",statsec=n_statsec,logfile=l_logfile)
date | scan(n_struct)
gemhedit(n_ref//"_avg","GEM-TLM",n_struct,
  "Last modification with GEMINI")
gemhedit(n_ref//"_avg","IMCOADD",n_struct,"Time stamp for imcoadd")
hedit(n_ref//"_avg","COMBSC","imcoadd combine not scaled with RELINT",add=yes,
  verify=no,show=no,update=yes,delete=no)
}
imarith(n_ref//"_avg","+",str(n_medsky),n_ref//"_avg",
  pixtype="real")
gemhedit(n_ref//"_avg","MED_SKY",str(n_medsky),"Median sky level")
# remove reference to BPM for reference image
hedit(n_ref//"_avg","BPM",delete=yes,verify=no,show=no,update=yes,add=no)

# Input ron and gain
# Update gain and readnoise keywords
imgets(n_ref//"_avg","GAINORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_avg","GAINORIG",n_gain[1],"Input gain")
}
imgets(n_ref//"_avg","RONORIG", >& "dev$null")
if (imgets.value=="0") {
   gemhedit(n_ref//"_avg","RONORIG",(sqrt(n_ron[1])*n_gain[1]),"Input read-noise")
}

# Effective ron and gain for combined image
# ron in e-/pix  as expected by other iraf tasks
# n_ron before this statement is RON^2 in ADU/pix
if(n_key_gain=="" || n_key_gain==" ")
  n_key_gain="GAIN"
gemhedit(n_ref//"_avg",n_key_gain,gaineff,"Effective gain [e-/ADU]")

if(n_key_ron=="" || n_key_ron==" ")
  n_key_ron="RDNOISE"
gemhedit(n_ref//"_avg",n_key_ron,(sqrt(roneff)),"Effective read-noise [e-]")

}  # end of fl_avg

printlog("-----------------------------------------------------------------------------",l_logfile,l_verbose)

# Delete temporary masks from DQ
if(n_immasks=="DQ" && n_fl_msk) 
  imdelete("@"//tmpmli,verify=no, >>& "dev$null")

delete(tmplis//","//tmpmli,verify=no, >>& "dev$null")

# update output headers to contain input images
# tmpout contains  image_trn of all images
if(n_fl_med || n_fl_add || n_fl_avg) {
  n_i=1
  inimag=tmpout
  while(fscan(inimag,n_inim)!=EOF) {
   if(n_i<10) {
     if(n_fl_med)
       gemhedit(n_ref//"_med","IMAGE0"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
     if(n_fl_add)
       gemhedit(n_ref//"_add"//l_zero_ext,"IMAGE0"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
     if(n_fl_avg)
       gemhedit(n_ref//"_avg","IMAGE0"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
   } else {
     if(n_fl_med)
       gemhedit(n_ref//"_med","IMAGE"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
     if(n_fl_add)
       gemhedit(n_ref//"_add"//l_zero_ext,"IMAGE"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
     if(n_fl_avg)
       gemhedit(n_ref//"_avg","IMAGE"//str(n_i),
         substr(n_inim,1,strlen(n_inim)-4),"Input image for imcoadd")
   }
   n_i+=1
  }
  inimag=""
  
  delete(tmpout,verify=no, >>& "dev$null")
}

end
